﻿using Microsoft.SqlServer.Server;
using System;
using System.Data.SqlTypes;
using System.IO;
using System.Text;
using System.Xml;
using System.Xml.Schema;
using System.Xml.Serialization;

namespace MiLambda.Person
{
    #region LegalPerson
    /// <summary>
    /// Legal person class. Can be used as a SQL CLR User-defined type.
    /// XML attributes (XmlRoot and XmlAttribute) are defined to allow XML serialisation
    /// in accordance with the supplied XML Schema definitions.
    /// </summary>
    [Serializable]
    [SqlUserDefinedType
        (
        Format.UserDefined,
        IsByteOrdered = true,
        IsFixedLength = false,
        MaxByteSize = 2048,
        Name = "LegalPersonClr"
        )]
    [XmlRoot
        (
        "legalPerson",
        Namespace = PersonInfo.LEGAL_PERSON_NAMESPACE,
        IsNullable = false
        )]
    public class LegalPerson : INullable, IBinarySerialize
    {
        #region Private Members
        private Boolean _isNull;
        #endregion

        #region Public Members
        [SqlFacet
            (
            IsFixedLength = false,
            IsNullable = false,
            MaxSize = 450
            )]
        [XmlAttribute
            (
            "fullName",
            Form = XmlSchemaForm.Qualified
            )]
        public String FullName { get; set; }

        [SqlFacet
            (
            IsFixedLength = false,
            IsNullable = false,
            MaxSize = 450
            )]
        [XmlAttribute
            (
            "abbreviatedName",
            Form = XmlSchemaForm.Qualified
            )]
        public String AbbreviatedName { get; set; }

        [XmlAttribute
            (
            "registrationDate",
            Form = XmlSchemaForm.Qualified,
            DataType = "date"
            )]
        public DateTime RegistrationDate { get; set; }

        [SqlFacet
            (
            IsFixedLength = false,
            IsNullable = false,
            MaxSize = 15
            )]
        [XmlAttribute
            (
            "companyNumber",
            Form = XmlSchemaForm.Qualified
            )]
        public String CompanyNumber { get; set; }
        #endregion

        #region INullable Implementation
        public Boolean IsNull
        {
            get
            {
                return _isNull;
            }
        }

        public static LegalPerson Null
        {
            get
            {
                LegalPerson legalPerson = new LegalPerson();
                legalPerson._isNull = true;
                return legalPerson;
            }
        }
        #endregion

        #region Constructors
        /// <summary>
        /// Empty constructor (required by the SQL CLR UDT standard).
        /// </summary>
        public LegalPerson()
        {
        }

        /// <summary>
        /// Complete constructor.
        /// </summary>
        /// <param name="fullName"></param>
        /// <param name="abbreviatedName"></param>
        /// <param name="registrationDate"></param>
        /// <param name="companyNumber"></param>
        public LegalPerson(String fullName, String abbreviatedName, DateTime registrationDate, String companyNumber)
        {
            this.FullName = fullName;
            this.AbbreviatedName = abbreviatedName;
            this.RegistrationDate = registrationDate;
            this.CompanyNumber = companyNumber;
        }

        /// <summary>
        /// Custom constructor (allows the instance to be created from its XML
        /// representation).
        /// </summary>
        /// <param name="legalPersonXml"></param>
        public LegalPerson(XmlDocument legalPersonXml)
        {
            if (legalPersonXml.DocumentElement.NamespaceURI == PersonInfo.LEGAL_PERSON_NAMESPACE)
            {
                legalPersonXml.Schemas.Add
                    (
                    PersonInfo.CreateXmlSchemaSet
                        (
                        PersonInfo.LEGAL_PERSON_XML_SCHEMA_FILE_NAME,
                        PersonInfo.LEGAL_PERSON_NAMESPACE
                        )
                    );

                LegalPerson legalPerson = (LegalPerson)PersonInfo.CreateInstance
                    (
                    legalPersonXml,
                    this.GetType()
                    );

                this.FullName = legalPerson.FullName;
                this.AbbreviatedName = legalPerson.AbbreviatedName;
                this.RegistrationDate = legalPerson.RegistrationDate;
                this.CompanyNumber = legalPerson.CompanyNumber;
            }
            else
            {
                this._isNull = true;
            }
        }
        #endregion

        #region ToString (override)
        /// <summary>
        /// Returns legal person XML document string representation.
        /// </summary>
        /// <returns></returns>
        [SqlMethod
            (
            DataAccess = DataAccessKind.None,
            IsDeterministic = true,
            IsPrecise = true,
            OnNullCall = false
            )]
        public override String ToString()
        {
            return Encoding.Unicode.GetString(PersonInfo.CreateStream(this).ToArray());
        }
        #endregion

        #region Parse
        /// <summary>
        /// Creates an instance from its XML document string representation.
        /// </summary>
        /// <param name="legalPersonXmlString"></param>
        /// <returns></returns>
        [SqlMethod(OnNullCall = false)]
        public static LegalPerson Parse([SqlFacet(MaxSize = -1)] SqlString legalPersonXmlString)
        {
            if (legalPersonXmlString.IsNull)
            {
                return Null;
            }

            XmlReaderSettings settings = new XmlReaderSettings();
            settings.IgnoreWhitespace = true;

            XmlDocument legalPersonXml = new XmlDocument();
            using (XmlReader reader = XmlReader.Create(new MemoryStream(Encoding.UTF8.GetBytes(legalPersonXmlString.Value)), settings))
            {
                legalPersonXml.Load(reader);
                reader.Close();
            }

            return new LegalPerson(legalPersonXml);
        }
        #endregion

        #region Public Methods
        /// <summary>
        /// Returns the FriendlyName property as a deterministic function.
        /// </summary>
        /// <returns></returns>
        [SqlMethod
            (
            DataAccess = DataAccessKind.None,
            IsDeterministic = true,
            IsPrecise = true,
            OnNullCall = false
            )]
        public SqlString FriendlyName()
        {
            return new SqlString(this.AbbreviatedName);
        }

        /// <summary>
        /// Returns the CompanyNumber property as a deterministic function.
        /// </summary>
        /// <returns></returns>
        [SqlMethod
            (
            DataAccess = DataAccessKind.None,
            IsDeterministic = true,
            IsPrecise = true,
            OnNullCall = false
            )]
        public SqlString CompanyNumberIndex()
        {
            return new SqlString(this.CompanyNumber);
        }
        #endregion

        #region IBinarySerialize Implementation
        /// <summary>
        /// The custom binary serialisation implementation is needed whenever the SQL CLR
        /// UDT implements members of non-primitive types (for instance, of type String).
        /// </summary>
        void IBinarySerialize.Read(BinaryReader reader)
        {
            LegalPerson legalPerson = new LegalPerson(PersonInfo.CreateXmlDocument(reader.ReadString()));
            this.FullName = legalPerson.FullName;
            this.AbbreviatedName = legalPerson.AbbreviatedName;
            this.RegistrationDate = legalPerson.RegistrationDate;
            this.CompanyNumber = legalPerson.CompanyNumber;
        }

        void IBinarySerialize.Write(BinaryWriter writer)
        {
            writer.Write(this.ToString());
        }
        #endregion
    }
    #endregion

    #region NaturalPerson
    /// <summary>
    /// Natural person class. Can be used as a SQL CLR User-defined type.
    /// XML attributes (XmlRoot and XmlAttribute) are defined to allow XML serialisation
    /// in accordance with the supplied XML Schema definitions.
    /// </summary>
    [Serializable]
    [SqlUserDefinedType
        (
        Format.UserDefined,
        IsByteOrdered = true,
        IsFixedLength = false,
        MaxByteSize = 2048,
        Name = "LegalPersonClr"
        )]
    [XmlRoot
        (
        "naturalPerson",
        Namespace = PersonInfo.NATURAL_PERSON_NAMESPACE,
        IsNullable = false
        )]
    public class NaturalPerson : INullable, IBinarySerialize
    {
        #region Private Constants
        String __naturalPersonFriendlyNamePattern = "{0}, {1} {2}";
        #endregion

        #region Private Members
        private Boolean _isNull;
        #endregion

        #region Public Members
        [SqlFacet
            (
            IsFixedLength = false,
            IsNullable = false,
            MaxSize = 50
            )]
        [XmlAttribute
            (
            "firstName",
            Form = XmlSchemaForm.Qualified
            )]
        public String FirstName { get; set; }

        [SqlFacet
            (
            IsFixedLength = false,
            IsNullable = true,
            MaxSize = 50
            )]
        [XmlAttribute
            (
            "middleName",
            Form = XmlSchemaForm.Qualified
            )]
        public String MiddleName { get; set; }

        [SqlFacet
            (
            IsFixedLength = false,
            IsNullable = false,
            MaxSize = 50
            )]
        [XmlAttribute
            (
            "lastName",
            Form = XmlSchemaForm.Qualified
            )]
        public String LastName { get; set; }

        [XmlAttribute
            (
            "birthDate",
            Form = XmlSchemaForm.Qualified,
            DataType = "date"
            )]
        public DateTime BirthDate { get; set; }

        [SqlFacet
            (
            IsFixedLength = false,
            IsNullable = false,
            MaxSize = 12
            )]
        [XmlAttribute
            (
            "ssn",
            Form = XmlSchemaForm.Qualified
            )]
        public String SSN { get; set; }
        #endregion

        #region INullable Implementation
        public Boolean IsNull
        {
            get
            {
                return _isNull;
            }
        }

        public static NaturalPerson Null
        {
            get
            {
                NaturalPerson naturalPerson = new NaturalPerson();
                naturalPerson._isNull = true;
                return naturalPerson;
            }
        }
        #endregion

        #region Constructors
        /// <summary>
        /// Empty constructor (required by the SQL CLR UDT standard).
        /// </summary>
        public NaturalPerson()
        {
        }

        /// <summary>
        /// Complete constructor.
        /// </summary>
        /// <param name="firstName"></param>
        /// <param name="lastName"></param>
        /// <param name="birthDate"></param>
        /// <param name="ssn"></param>
        /// <param name="middleName"></param>
        public NaturalPerson(String firstName, String lastName, DateTime birthDate, String ssn, String middleName = null)
        {
            this.FirstName = firstName;
            this.MiddleName = middleName;
            this.LastName = lastName;
            this.BirthDate = birthDate;
            this.SSN = ssn;
        }

        /// <summary>
        /// Custom constructor (allows the instance to be created from its XML
        /// representation).
        /// </summary>
        /// <param name="naturalPersonXml"></param>
        public NaturalPerson(XmlDocument naturalPersonXml)
        {
            if (naturalPersonXml.DocumentElement.NamespaceURI == PersonInfo.NATURAL_PERSON_NAMESPACE)
            {
                naturalPersonXml.Schemas.Add
                    (
                    PersonInfo.CreateXmlSchemaSet
                        (
                        PersonInfo.NATURAL_PERSON_XML_SCHEMA_FILE_NAME,
                        PersonInfo.NATURAL_PERSON_NAMESPACE
                        )
                    );

                NaturalPerson naturalPerson = (NaturalPerson)PersonInfo.CreateInstance
                    (
                    naturalPersonXml,
                    this.GetType()
                    );

                this.FirstName = naturalPerson.FirstName;
                this.MiddleName = naturalPerson.MiddleName;
                this.LastName = naturalPerson.LastName;
                this.BirthDate = naturalPerson.BirthDate;
                this.SSN = naturalPerson.SSN;
            }
            else
            {
                this._isNull = true;
            }
        }
        #endregion

        #region ToString (override)
        /// <summary>
        /// Returns natural person XML document string representation.
        /// </summary>
        /// <returns></returns>
        [SqlMethod
            (
            DataAccess = DataAccessKind.None,
            IsDeterministic = true,
            IsPrecise = true,
            OnNullCall = false
            )]
        public override String ToString()
        {
            return Encoding.Unicode.GetString(PersonInfo.CreateStream(this).ToArray());
        }
        #endregion

        #region Parse
        /// <summary>
        /// Creates an instance from its XML document string representation.
        /// </summary>
        /// <param name="naturalPersonXmlString"></param>
        /// <returns></returns>
        [SqlMethod(OnNullCall = false)]
        public static NaturalPerson Parse([SqlFacet(MaxSize = -1)] SqlString naturalPersonXmlString)
        {
            if (naturalPersonXmlString.IsNull)
            {
                return Null;
            }

            XmlReaderSettings settings = new XmlReaderSettings();
            settings.IgnoreWhitespace = true;

            XmlDocument naturalPersonXml = new XmlDocument();
            using (XmlReader reader = XmlReader.Create(new MemoryStream(Encoding.UTF8.GetBytes(naturalPersonXmlString.Value)), settings))
            {
                naturalPersonXml.Load(reader);
                reader.Close();
            }

            return new NaturalPerson(naturalPersonXml);
        }
        #endregion

        #region Public Methods
        /// <summary>
        /// Returns the FriendlyName property as a deterministic function.
        /// </summary>
        /// <returns></returns>
        [SqlMethod
            (
            DataAccess = DataAccessKind.None,
            IsDeterministic = true,
            IsPrecise = true,
            OnNullCall = false
            )]
        public SqlString FriendlyName()
        {
            String middleName = this.MiddleName;

            SqlString result = new SqlString
                (
                String.Format
                    (
                    __naturalPersonFriendlyNamePattern,
                    this.LastName,
                    this.FirstName,
                    (String.IsNullOrEmpty(middleName) ? String.Empty : middleName)
                    ).Trim()
                );
            return result;
        }

        /// <summary>
        /// Returns the SSN property as a deterministic function.
        /// </summary>
        /// <returns></returns>
        [SqlMethod
            (
            DataAccess = DataAccessKind.None,
            IsDeterministic = true,
            IsPrecise = true,
            OnNullCall = false
            )]
        public SqlString SSNIndex()
        {
            return new SqlString(this.SSN);
        }
        #endregion

        #region IBinarySerialize Implementation
        /// <summary>
        /// The custom binary serialisation implementation is needed whenever the SQL CLR
        /// UDT implements members of non-primitive types (for instance, of type String).
        /// </summary>
        void IBinarySerialize.Read(BinaryReader reader)
        {
            NaturalPerson naturalPerson = new NaturalPerson(PersonInfo.CreateXmlDocument(reader.ReadString()));
            this.FirstName = naturalPerson.FirstName;
            this.MiddleName = naturalPerson.MiddleName;
            this.LastName = naturalPerson.LastName;
            this.BirthDate = naturalPerson.BirthDate;
            this.SSN = naturalPerson.SSN;
        }

        void IBinarySerialize.Write(BinaryWriter writer)
        {
            writer.Write(this.ToString());
        }
        #endregion
    }
    #endregion
}
