﻿using System;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Xml;
using System.Xml.Schema;
using System.Xml.Serialization;

namespace MiLambda.Person
{
    #region PersonInfo
    /// <summary>
    /// The PersonInfo abstract class contains common features shared by
    /// MiLambda.Person.LegalPerson, and MiLambda.Person.NaturalPerson, classes.
    /// </summary>
    public abstract class PersonInfo
    {
        #region Public Constants
        /// <summary>
        /// Xml settings.
        /// </summary>
        public const String PERSON_NAMESPACE = "http://milambda.net/person";
        public const String LEGAL_PERSON_NAMESPACE = "http://milambda.net/legalPerson";
        public const String LEGAL_PERSON_XML_SCHEMA_FILE_NAME = "MiLambda.LegalPerson.xsd";
        public const String NATURAL_PERSON_NAMESPACE = "http://milambda.net/naturalPerson";
        public const String NATURAL_PERSON_XML_SCHEMA_FILE_NAME = "MiLambda.NaturalPerson.xsd";

        /// <summary>
        /// Messages.
        /// </summary>
        public const String UNKNOWN_XML_DOCUMENT_MESSAGE = "Unknown XML Document.";
        #endregion

        #region Private Constants
        /// <summary>
        /// Sometimes it's hard to find the begining of the XML inside an XML string...
        /// </summary>
        private const String __beginsWith = "<";
        #endregion

        #region Internal Static Methods
        /// <summary>
        /// Creates memory stream from object instance.
        /// </summary>
        /// <param name="instance"></param>
        /// <returns></returns>
        internal static MemoryStream CreateStream(Object instance)
        {
            using (MemoryStream stream = new MemoryStream())
            {
                XmlWriterSettings settings = new XmlWriterSettings();
                settings.Encoding = Encoding.Unicode;
                settings.OmitXmlDeclaration = true;

                XmlSerializer serializer = new XmlSerializer(instance.GetType());
                using (XmlWriter writer = XmlWriter.Create(stream, settings))
                {
                    serializer.Serialize(writer, instance);
                    writer.Close();
                }

                stream.Close();

                return stream;
            }
        }

        /// <summary>
        /// Creates XML document from its string representation.
        /// </summary>
        /// <param name="instanceXmlString"></param>
        /// <returns></returns>
        internal static XmlDocument CreateXmlDocument(String instanceXmlString)
        {
            XmlDocument naturalPersonXml = new XmlDocument();
            /// For some perverse reason, the XML string generated by SQL Server can only
            /// be parsed safely, if you "manually" extract the XML definition from the
            /// string.
            naturalPersonXml.LoadXml(instanceXmlString.Substring(instanceXmlString.IndexOf(__beginsWith)));
            return naturalPersonXml;
        }

        /// <summary>
        /// Creates object instance from its XML document representation.
        /// </summary>
        /// <param name="instanceXml"></param>
        /// <param name="type"></param>
        /// <returns></returns>
        internal static Object CreateInstance(XmlDocument instanceXml, Type type)
        {
            if (instanceXml.Schemas.Count > 0)
            {
                instanceXml.Validate(null);
            }

            XmlReaderSettings settings = new XmlReaderSettings();
            settings.IgnoreProcessingInstructions = true;
            settings.IgnoreWhitespace = true;

            XmlSerializer serializer = new XmlSerializer(type);
            using (XmlReader reader = XmlReader.Create(new XmlNodeReader(instanceXml), settings))
            {
                return serializer.Deserialize(reader);
            }
        }

        /// <summary>
        /// Creates Xml Schema Set.
        /// </summary>
        /// <param name="xmlSchemaFileName"></param>
        /// <param name="xmlSchemaNamespace"></param>
        /// <returns></returns>
        internal static XmlSchemaSet CreateXmlSchemaSet(String xmlSchemaFileName, String xmlSchemaNamespace)
        {
            XmlSchemaSet xmlSchemaSet = new XmlSchemaSet();

            Assembly assembly = Assembly.GetExecutingAssembly();

            String xmlSchemaResourceName = assembly.GetManifestResourceNames().FirstOrDefault<String>
                (
                resourceName
                =>
                resourceName.EndsWith(xmlSchemaFileName)
                );

            if (!String.IsNullOrEmpty(xmlSchemaResourceName))
            {
                using (XmlReader reader = XmlReader.Create(assembly.GetManifestResourceStream(xmlSchemaResourceName)))
                {
                    xmlSchemaSet.Add(xmlSchemaNamespace, reader);
                    reader.Close();
                }
            }

            return xmlSchemaSet;
        }

        #endregion
    }
    #endregion
}
